﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using SimpleJSON; // JSONNode
using MVR.FileManagementSecure; // WriteAllBytes
using System.Text.RegularExpressions;

namespace JustAnotherUser {
    public class GenderSwapper : MVRScript {
        private static readonly string VERSION = "0.1";
        private static readonly int UUID_LENGTH = 8;

        public static readonly int DIFFUSE_TEXTURE = 0,
                                    SPECULAR_TEXTURE = 1,
                                    GLOSS_TEXTURE = 2,
                                    NORMAL_TEXTURE = 3,
                                    DECAL_TEXTURE = 5;

        //private DecalMakerHelper _helper;
        private CUAManagerPath _path;

        private DAZCharacterSelector _dazCharacterSelector;
        private JSONStorable _decalmakerStorable;
        private List<string> _decalmakerProgressStorables;
        private List<DAZMorph> _morphs;

        private PresetData _origin;
        private PresetData _destiny;

        private GenderSwapperUI _ui;

        private object _decalIsLoadingSync = new object();
        private bool _decalIsLoading = false;

        private string _baseName;

        public override void Init() {
            // plugin VaM GUI description
            pluginLabelJSON.val = "GenderSwapper v" + VERSION;

            this._path = new CUAManagerPath(this, "Saves\\PluginData\\mrmr32\\GenderSwapper");
            this._path.VerifyPluginDataFolder();
            this._path.Init();

            this._dazCharacterSelector = containingAtom.GetComponentInChildren<DAZCharacterSelector>();
            if (this._dazCharacterSelector == null) {
                SuperController.LogError("You need to aply GenderSwapper on a Person atom");
                return;
            }
            this._decalmakerProgressStorables = new List<string>();
            this._morphs = MorphHelper.ScanBank(this._dazCharacterSelector, ignoreGender: true);

            this._origin = new PresetData();
            this._destiny = new PresetData();

            this._ui = new GenderSwapperUI(this);
            this._ui.SetupUI();
            this._ui.SetOnSliderChange(SetProgress);
            this._ui.SetLoadPresetButtonEvent((val) => {
                try {
                    this._origin.RestoreFromAppearancePreset(GetFileName(val,"vap"), SuperController.singleton.LoadJSON(val).AsObject);
                    this.LoadData();
                } catch (Exception ex) {
                    SuperController.LogError(ex.ToString());
                }
            }, false);
            this._ui.SetLoadPresetButtonEvent((val) => {
                try {
                    this._destiny.RestoreFromAppearancePreset(GetFileName(val, "vap"), SuperController.singleton.LoadJSON(val).AsObject);
                    this.LoadData();
                } catch (Exception ex) {
                    SuperController.LogError(ex.ToString());
                }
            }, true);

            // TODO add scene name instead of random
            this._baseName = GetRandomName(UUID_LENGTH) + "_" + containingAtom.name;
        }

        protected void Start() {
            string decalMakerId = this.GetDecalMaker();
            if (decalMakerId == null) {
                SuperController.LogError("GenderSwapper needs DecalMaker in order to work. Please, first add DecalMaker.");
                return;
            }
            this._decalmakerStorable = containingAtom.GetStorableByID(decalMakerId);

            UVData.Load();

            StartCoroutine(LoadDataOnceAllFinishes());
        }

        private IEnumerator LoadDataOnceAllFinishes() {
            // wait the UVData to finish loading
            yield return new WaitUntil(() => UVData.IsLoaded());

            this.LoadData(retryIfFail: true); // DecalMaker may fail at the beginning
        }

        private string GetDecalMaker() {
            Regex decalName = new Regex(@"plugin#\d+_VAM_Decal_Maker.Decal_Maker");
            foreach (string id in containingAtom.GetStorableIDs()) {
                if (decalName.Match(id).Success) return id;
            }

            return null; // none
        }

        public void SetProgress(float val) {
            if (!this._origin.isValid || !this._destiny.isValid) return; // we need both
            
            Dictionary<string,float> result = LeapMorphs(this.GetMorphs(false), this.GetMorphs(true), val);
            foreach (var entry in result) {
                MorphHelper.SetMorphValue(MorphHelper.FindMorphByName(this._morphs, entry.Key), entry.Value);
            }

            if (this._decalmakerStorable == null) return;
            lock (this._decalIsLoadingSync) {
                foreach (string storable in this._decalmakerProgressStorables) {
                    this._decalmakerStorable.SetFloatParamValue(storable, 1.0f - val); // as we're assigning the source here, 0% progress is 100% source, and 100% progress 0% source
                }
            }
        }

        /**
         * Loads the `origin` texture into the first slot of DecalMaker (if no slot is present, a new one will be placed).
         * @param decalMaker:   JSONClass that contains decalMaker. This information will be changed
         * @param origin:       Textures to load. This must be of the same sex as the atom.
         **/
        public static void LoadOriginTexturesIntoDecalMaker(JSONClass decalMaker, AtomTexture origin, float progress = 0f) {
            if (!origin.isValid) return;
            int decalVersion = decalMaker["SaveVersion"].AsInt;
            LoadOriginTexturesIntoDecalMaker(decalMaker, decalVersion, "face", origin.head, progress);
            LoadOriginTexturesIntoDecalMaker(decalMaker, decalVersion, "torso", origin.torso, progress);
            LoadOriginTexturesIntoDecalMaker(decalMaker, decalVersion, "limbs", origin.limbs, progress);
        }

        public static void LoadOriginTexturesIntoDecalMaker(JSONClass decalMaker, int decalVersion, string part, Texture originTextureOfPart, float progress) {
            if (originTextureOfPart.diffuse.Length > 0) LoadOriginTexturesIntoDecalMaker(decalMaker, "_DecalTex", part, originTextureOfPart.diffuse, progress);
            if (originTextureOfPart.specular.Length > 0) LoadOriginTexturesIntoDecalMaker(decalMaker, "_SpecTex", part, originTextureOfPart.specular, progress);
            if (originTextureOfPart.gloss.Length > 0) LoadOriginTexturesIntoDecalMaker(decalMaker, "_GlossTex", part, originTextureOfPart.gloss, progress);
            if (originTextureOfPart.normal.Length > 0) LoadOriginTexturesIntoDecalMaker(decalMaker, "_BumpMap", part, originTextureOfPart.normal, progress);
            // TODO add decal as another row
        }

        public static void LoadOriginTexturesIntoDecalMaker(JSONClass decalMaker, string cls, string part, string texturePath, float progress) {
            JSONArray elementsInPart = decalMaker[cls][part].AsArray;

            JSONClass texture = (elementsInPart.Count > 0 ? elementsInPart[0].AsObject : null);
            if (texture == null) {
                texture = new JSONClass();
                elementsInPart[0] = texture;
                texture["H"].AsFloat = 0f;
                texture["S"].AsFloat = 0f;
                texture["V"].AsFloat = 1f;
                texture["RandomName"] = GetRandomName(7);
                texture["LinkID"] = "*";
            }

            texture["Alpha"].AsFloat = 1.0f - progress; // as we're assigning the source here, 0% progress is 100% source, and 100% progress 0% source
            texture["Path"] = texturePath;
        }

        private void GetDecalMakerStorablesIds(JSONClass decalMaker) {
            List<string> storables = new List<string>();
            // TODO access Decal  to change its value; load all the names to modify on the `LoadDecalMaker`

            string prefix = "Alpha";
            // @ref got from BodyRegionEnum and MatSlotEnum
            string MaterialSlot = "_DecalTex"; // only decal has alpha value
            foreach (string TextureSlot in new String[] { "torso", "face", "limbs" }) {
                JSONArray textures = decalMaker[MaterialSlot][TextureSlot].AsArray;
                if (textures.Count > 0) {
                    // we have one slider
                    string RandomName = textures[0].AsObject["RandomName"];
                    string sliderId = string.Format("{0}_{1}{2}{3}", prefix, TextureSlot, MaterialSlot, RandomName); // @ref from CreateJSN
                    storables.Add(sliderId);
                }
            }

            this._decalmakerProgressStorables = storables;
        }

        private static string GetFileName(string path, string prefix) {
            var pattern = @"(Preset_)?([^\\/]+)\." + prefix;
            var match = Regex.Match(path, pattern);
            return match.Groups[2].Value;
        }
        
		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false) {
            JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
            if (includePhysical || forceStore) {
                jc["origin"] = this._origin.GetJSON(this.subScenePrefix);
                jc["destiny"] = this._destiny.GetJSON(this.subScenePrefix);
            }
            return jc;
        }

        public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical, bool restoreAppearance, bool setMissingToDefault) {
            base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);

            if (!this.physicalLocked && restorePhysical) {
                this._origin = new PresetData();
                this._destiny = new PresetData();

                this._origin.RestoreFromJSON(jc["origin"].AsObject, this.subScenePrefix, this.mergeRestore, setMissingToDefault);
                this._destiny.RestoreFromJSON(jc["destiny"].AsObject, this.subScenePrefix, this.mergeRestore, setMissingToDefault);

                this.LoadData(); // now that we have the origin/target, update the UI
            }
        }

        /**
         * From origin's morphs to target's. If no morph is present at its counterpart, then a value of 0 is assumed.
         * TODO shouldn't we keep two different dictionaries for male-female morphs?
         **/
        public Dictionary<string, float> LeapMorphs(Dictionary<string, float> origin, Dictionary<string, float> target, float targetPercentage) {
            Dictionary<string, float> r = new Dictionary<string, float>();
            
            foreach (string morph in origin.Keys.Union(target.Keys)) {
                float from = (origin.ContainsKey(morph) ? origin[morph] : 0f),
                    to = (target.ContainsKey(morph) ? target[morph] : 0f);

                // 1 to 0 at 0% is 1, at 50% 0.5, at 100% 0
                // 0 to 1 at 0% is 0, at 50% 0.5, at 100% 1
                // 1 to 1 is always 1
                float diff = to - from;
                float current = from + diff*targetPercentage;

                r[morph] = current;
            }

            return r;
        }


        private void LoadData(bool retryIfFail = false) {
            this.LoadUIData();

            if (this._destiny.isValid) {
                // load destiny data (eyes, sex...)
                bool useOtherSexMorphs = (this._origin.isValid && this._destiny.isValid) && (this._origin.isMale != this._destiny.isMale);

                JSONClass geometry = this._dazCharacterSelector.GetJSON();
                this._destiny.LoadGeometryToAtom(geometry, useOtherSexMorphs);
                this._dazCharacterSelector.RestoreFromJSON(geometry);

                foreach (var morph in this._destiny.morphs) MorphHelper.SetMorphValue(MorphHelper.FindMorphByName(this._morphs, morph.Key), morph.Value); // LoadToAtom don't load the morphs
                
                DAZCharacterTextureControl t = this.containingAtom.GetStorableByID("textures") as DAZCharacterTextureControl;
                JSONClass textures = t.GetJSON();
                this._destiny.LoadTexturesToAtom(textures);
                t.RestoreFromJSON(textures); 
                    
                if (this._decalmakerStorable != null) this._decalmakerStorable.CallAction("Reset To Original Textures");
            }

            // update morphs %
            this.SetProgress(this._ui.GetSliderProgress());

            // load the origin decal texture
            StartCoroutine(this.LoadDecalMaker(this._ui.GetSliderProgress(), retryIfFail));
        }
        
        private IEnumerator LoadDecalMaker(float progress = 0f, bool retryIfFail = false) {
            bool isYours = false;
            lock (this._decalIsLoadingSync) {
                isYours = !this._decalIsLoading;
                if (isYours) this._decalIsLoading = true;
            }
            if (!isYours) yield break; // already loading

            Exception exception = null; // TODO try/catch, saving the exception (if any) here
            if (this._origin.isValid && this._destiny.isValid && this._decalmakerStorable != null) { // we need destiny data to create the origin decal
                JSONClass cls = new JSONClass();
                try {
                    this._decalmakerStorable.GetJSON();
                } catch (Exception ex) {
                    exception = ex;
                }

                if (exception == null) { // continue?
                    AtomTexture texture = (this._destiny.isMale ? this._origin.maleTexture : this._origin.femaleTexture);
                    if (texture == null) {
                        StartCoroutine(LoadTexture(this._origin, this._decalIsLoadingSync, this._destiny.isMale, this._path.PluginDataFolder, this._baseName));

                        // wait `LoadTexture` to finish
                        yield return new WaitUntil(() => {
                            lock (this._decalIsLoadingSync) {
                                return (this._destiny.isMale ? this._origin.maleTexture : this._origin.femaleTexture) != null;
                            }
                        });

                        texture = (this._destiny.isMale ? this._origin.maleTexture : this._origin.femaleTexture); // get the loaded texture
                    }

                    try {
                        // @pre the texture must be the same sex as the atom
                        LoadOriginTexturesIntoDecalMaker(cls, texture, progress);
                        // TODO give the user a config to set 'Nipple Cutouts ON' to false

                        this._decalmakerStorable.CallAction("ClearAll"); //removes any existing loaded decals
                        this._decalmakerStorable.RestoreFromJSON(cls);
                        this._decalmakerStorable.CallAction("PerformLoad"); //load the data from the _savedData variable

                        GetDecalMakerStorablesIds(cls);
                    } catch (Exception ex) {
                        exception = ex;
                    }
                }
            }

            lock (this._decalIsLoadingSync) {
                this._decalIsLoading = false;
            }

            if (exception != null) {
                // failed
                SuperController.LogError(exception.ToString());
                if (retryIfFail) {
                    // try again
                    yield return new WaitForSeconds(0.5f);
                    StartCoroutine(this.LoadDecalMaker(progress, retryIfFail));
                }
            }
        }

        private void LoadUIData() {
            if (this._origin.isValid) {
                this._ui.SetName(this._origin.name + (this._origin.isMale ? " (male)" : " (female)"), false);
                // TODO mark body&iris as invalid
                this._ui.SetBody(this._origin.skin, false);
                this._ui.SetIris(this._origin.iris, false);
                this._ui.SetMorphList(this.MorphListToString(false), false);
                this._ui.SetTextureList(this._origin.texture.ToString(), false); // TODO show also converted texture path
            }

            if (this._destiny.isValid) {
                this._ui.SetName(this._destiny.name + (this._destiny.isMale ? " (male)" : " (female)"), true);
                this._ui.SetBody(this._destiny.skin, true);
                this._ui.SetIris(this._destiny.iris, true);
                this._ui.SetMorphList(this.MorphListToString(true), true);
                this._ui.SetTextureList(this._destiny.texture.ToString(), true);
            }
        }

        private string MorphListToString(bool target) {
            Dictionary<string, float> list = this.GetMorphs(target);

            string r = "";
            bool femaleConversion = false;
            foreach (var e in list) {
                if (e.Key.Equals("MVR_G2Female")) femaleConversion = true; // we'll place it at the bottom
                else r += e.Key + ": " + e.Value.ToString("0.00") + "\n";
            }

            if (femaleConversion) {
                r += "MVR_G2Female: " + list["MVR_G2Female"].ToString("0.00") + "*\n\n"
                    + "*As you’re changing from " + (this._origin.isMale ? "male" : "female") + " to " + (this._destiny.isMale ? "male" : "female") + " this morph needs to be added";
            }

            return r;
        }

        private Dictionary<string, float> GetMorphs(bool target) {
            if (target) return this._destiny.morphs;

            // the source may need to add the `MVR_G2Female` morph
            if (this._destiny.isMale == this._origin.isMale) return this._origin.morphs; // same sex; we don't need to convert anything

            Dictionary<string, float> r = new Dictionary<string, float>(this._origin.morphs);
            r["MVR_G2Female"] = (this._destiny.isMale ? 1f : -1f); // if it's a female converting into a male we need to set it on 1, if it's a male into a female set it to -1
            return r;
        }

        public static string GetRandomName(int lenght) {
            var chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789";
            string r = "";
            for (int i = 0; i < lenght; i++) r += chars[(int)UnityEngine.Random.Range(0, chars.Length)];
            return r;
        }

        private static IEnumerator LoadTexture(PresetData textures, object changingTexturesLock, bool intoMale, string folder, string baseOutFileName) {
            AtomTexture textureToTransform = textures.texture;
            AtomTexture transformedTexture = new AtomTexture();
            transformedTexture.head = new Texture();
            transformedTexture.torso = new Texture();
            transformedTexture.limbs = new Texture();

            // we'll have to convert as many textures as `AtomTexture` holds
            foreach (KeyValuePair<Texture, int> texture in textureToTransform.GetTextures()) {
                int region = texture.Value;
                foreach (KeyValuePair<string, int> tex in texture.Key.GetTextures()) {
                    string texturePath = tex.Key;
                    int typeOfTexture = tex.Value;


                    Texture2D originalTexture = new Texture2D(1, 1); // it will be changed later by `ImageConversion.LoadImage`

                    try {
                        SuperController.LogMessage("Converting image '" + texturePath + "'...");

                        // get the new image
                        byte[] originalData = FileManagerSecure.ReadAllBytes(texturePath);
                        ImageConversion.LoadImage(originalTexture, originalData);
                    } catch (Exception ex) {
                        SuperController.LogError(ex.ToString());
                    }

                    // wait the UVData to finish loading
                    yield return new WaitUntil(() => UVData.IsLoaded());

                    try {
                        // distort the image
                        Texture2D targetTexture = UVData.DeformUVs(originalTexture, !intoMale, region);

                        // get the output path
                        // @ref CUAManager
                        string filePath = CUAManagerPath.Combine(folder, baseOutFileName + "-" + region.ToString() + "_" + typeOfTexture.ToString() + ".png");
                        filePath = filePath.Replace("\\", "/"); // Sanitize input, just-in-case.
                
                        // convert the output image back into bytes
                        byte[] image = ImageConversion.EncodeToPNG(targetTexture);
                
                        // write the output image
                        FileManagerSecure.WriteAllBytes(filePath, image);
                
                        if (FileManagerSecure.FileExists(filePath)) {
                            SuperController.LogMessage("Image '" + texturePath + "' converted as '" + filePath + "'.");
                            transformedTexture.GetTexture(region).SetTexture(typeOfTexture, filePath);
                        }
                        else {
                            SuperController.LogError("Error while converting image '" + texturePath + "'");
                        }
                    } catch (Exception ex) {
                        SuperController.LogError(ex.ToString());
                    }
                }
            }

            
            
            lock (changingTexturesLock) {
                if (intoMale) textures.maleTexture = transformedTexture;
                else textures.femaleTexture = transformedTexture;
            }
        }
    }
}